from onegov.activity.utils import random_group_code
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.activity.models.occasion import Occasion
from sqlalchemy import Column, Enum, Index, Text, ForeignKey, Integer, func
from sqlalchemy.orm import object_session
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy_utils import aggregated
from uuid import uuid4


class Booking(Base, TimestampMixin):
    """ Bookings are created by users for occasions.

    Initially, bookings are open. In this state they represent a "wish" rather
    than booking.

    Because bookings are wishes initially, they get a priority as well as
    a group code which links multiple bookings by multiple users together.

    They system will try to figure out the best booking using the priority
    as well as the group code.

    """

    __tablename__ = 'bookings'

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    #: the public id of the booking
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the user owning the booking
    username = Column(Text, ForeignKey('users.username'), nullable=False)

    #: the priority of the booking, a higher number = a higher priority
    priority = Column(Integer, nullable=False, default=0)

    #: the group code of the booking
    group_code = Column(Text, nullable=False, default=random_group_code)

    #: the attendee behind this booking
    attendee_id = Column(UUID, ForeignKey("attendees.id"), nullable=False)

    #: the occasion this booking belongs to
    occasion_id = Column(UUID, ForeignKey("occasions.id"), nullable=False)

    #: the period this booking belongs to
    @aggregated('occasion', Column(
        UUID, ForeignKey("periods.id"), nullable=False)
    )
    def period_id(self):
        return func.coalesce(Occasion.period_id, None)

    #: the state of the booking
    state = Column(
        Enum(
            'open',
            'blocked',
            'accepted',
            'denied',
            'cancelled',
            name='booking_state'
        ),
        nullable=False,
        default='open'
    )

    __mapper_args__ = {
        'order_by': priority
    }

    __table_args__ = (
        Index(
            'one_booking_per_attendee', 'occasion_id', 'attendee_id',
            unique=True
        ),
        Index('bookings_by_state', 'state', 'username')
    )

    @hybrid_property
    def starred(self):
        return self.priority != 0

    def star(self, max_stars=3):
        """ Stars the current booking (giving it a priority of 1), up to
        a limit per period and attendee.

        :return: True if successful (or already set), False if over limit.

        """

        if self.starred:
            return True

        session = object_session(self)

        o = session.query(Occasion)
        o = o.with_entities(Occasion.id)
        o = o.filter(Occasion.period_id == self.occasion.period_id)

        q = session.query(Booking)
        q = q.filter(Booking.attendee_id == self.attendee_id)
        q = q.filter(Booking.username == self.username)
        q = q.filter(Booking.occasion_id.in_(o.subquery()))
        q = q.filter(Booking.id != self.id)
        q = q.filter(Booking.priority != 0)

        if q.count() < max_stars:
            self.priority = 1
            return True

        return False

    def unstar(self):
        self.priority = 0

    @property
    def score(self):
        """ The score of this booking for the matching algorithm. This should
        be a value that doesn't during the runtime of the algorithm (so
        no timestamp for example).

        """

        return self.priority

    @property
    def start(self):
        return self.occasion.start

    @property
    def end(self):
        return self.occasion.end



""" Implements the matching algorithm used to match attendees to occasions.

The algorithm used is based on Deferred Acceptance. The algorithm has a
quadratic runtime.

"""

from onegov.activity import Booking, Occasion
from onegov.activity.matching.utils import overlaps, LoopBudget, hashable
from onegov.core.utils import Bunch
from itertools import groupby, product
from sortedcontainers import SortedSet
from sqlalchemy.orm import joinedload


class AttendeeAgent(hashable('id')):
    """ Acts on behalf of the attendee with the goal to get a stable booking
    with an occasion.

    A booking/occasion pair is considered stable if there exists no other
    such pair which is preferred by both the attendee and the occasion.

    In other words, if there's no other occasion that would accept the
    attendee over another attendee.

    """

    __slots__ = ('id', 'wishlist', 'accepted', 'blocked')

    def __init__(self, id, bookings):
        self.id = id

        # keep the wishlist sorted from highest to lowest priority, ignoring
        # the current state of the booking as we will assign them anew
        # each time we run the algorithm
        def booking_order(booking):
            return booking.priority * -1, booking.id

        self.wishlist = SortedSet(bookings, key=booking_order)
        self.accepted = set()
        self.blocked = set()

    def accept(self, booking):
        """ Accepts the given booking. """

        self.wishlist.remove(booking)
        self.accepted.add(booking)

        self.blocked |= {b for b in self.wishlist if overlaps(booking, b)}
        self.wishlist -= self.blocked

    def deny(self, booking):
        """ Removes the given booking from the accepted bookings. """

        self.wishlist.add(booking)
        self.accepted.remove(booking)

        # remove bookings from the blocked list which are not blocked anymore
        free = set(self.blocked)

        for accepted in self.accepted:
            for blocked in self.blocked:
                if overlaps(accepted, blocked):
                    free.remove(blocked)

        self.blocked -= free
        self.wishlist |= free

    @property
    def is_valid(self):
        """ Returns True if the results of this agent are valid.

        The algorithm should never get to this stage, so this is an extra
        security measure to make sure there's no bug.

        """
        for a, b in product(self.accepted, self.accepted):
            if a != b and overlaps(a, b):
                return False

        return True


class OccasionAgent(hashable('id')):
    """ Represents the other side of the Attendee/Occasion pair.

    While the attende agent will try to get the best possible occasion
    according to the wishses of the attendee, the occasion agent will
    try to get the best attendee according to the wishes of the occasion.

    These wishes may include hard-coded rules or peferences defined by the
    organiser/admin, who may manually prefer certain attendees over others.

    """

    __slots__ = ('occasion', 'bookings', 'attendees')

    def __init__(self, occasion):
        self.id = occasion.id
        self.occasion = occasion
        self.bookings = set()
        self.attendees = {}

    @property
    def full(self):
        return len(self.bookings) >= (self.occasion.max_spots)

    def preferred(self, booking):
        """ Returns the first booking with a lower score than the given
        booking (which indicates that the given booking is preferred over
        the returned item).

        If there's no preferred booking, None is returned.

        """
        return next(
            (b for b in self.bookings if b.score < booking.score),
            None
        )

    def match(self, attendee, booking):

        # as long as there are spots, automatically accept new requests
        if not self.full:
            self.attendees[booking] = attendee
            self.bookings.add(booking)

            attendee.accept(booking)
            return True

        # if the occasion is already full, accept the booking by throwing
        # another one out, if there exists a better fit
        over = self.preferred(booking)

        if over:
            self.attendees[over].deny(over)
            self.bookings.remove(over)

            self.bookings.add(booking)
            attendee.accept(booking)

            return True

        return False


def deferred_acceptance_from_database(session, period_id, **kwargs):
    b = session.query(Booking)
    b = b.options(joinedload(Booking.occasion))
    b = b.filter(Booking.period_id == period_id)
    b = b.filter(Booking.state.in_(('open', 'accepted', 'blocked')))

    o = session.query(Occasion)
    o = o.filter(Occasion.period_id == period_id)

    bookings = deferred_acceptance(bookings=b, occasions=o, **kwargs)

    # write the changes to the database
    def update_states(bookings, state):
        ids = set(b.id for b in bookings)

        if not ids:
            return

        b = session.query(Booking)
        b = b.filter(Booking.id.in_(ids))

        for booking in b:
            booking.state = state

    update_states(bookings.open, 'open')
    update_states(bookings.accepted, 'accepted')
    update_states(bookings.blocked, 'blocked')


def deferred_acceptance(bookings, occasions,
                        validity_check=True,
                        stability_check=False,
                        hard_budget=False):
    """ Matches bookings with occasions. """

    bookings = [b for b in bookings]
    bookings.sort(key=lambda b: b.attendee_id)

    occasions = {
        o.id: OccasionAgent(o) for o in occasions
    }

    attendees = {
        aid: AttendeeAgent(aid, bookings)
        for aid, bookings in groupby(bookings, key=lambda b: b.attendee_id)
    }

    # I haven't proven yet that the following loop will always end. Until I
    # do there's a fallback check to make sure that we'll stop at some point
    budget = LoopBudget(max_ticks=len(bookings) * len(attendees))

    # while there are attendees with entries in a wishlist
    while next((a for a in attendees.values() if a.wishlist), None):

        if budget.limit_reached(as_exception=hard_budget):
            break

        candidates = [a for a in attendees.values() if a.wishlist]
        matched = 0

        # match attendees to courses
        while candidates:
            candidate = candidates.pop()

            for booking in candidate.wishlist:
                if occasions[booking.occasion_id].match(candidate, booking):
                    matched += 1
                    break  # required because the wishlist has been changed

        # if no matches were possible the situation can't be improved
        if not matched:
            break

    # make sure the algorithm didn't make any mistakes
    if validity_check:
        for a in attendees.values():
            assert a.is_valid

    # make sure the result is stable
    if stability_check:
        assert is_stable(attendees.values(), occasions.values())

    return Bunch(
        open=set(b for a in attendees.values() for b in a.wishlist),
        accepted=set(b for a in attendees.values() for b in a.accepted),
        blocked=set(b for a in attendees.values() for b in a.blocked)
    )


def is_stable(attendees, occasions):
    """ Returns true if the matching between attendees and occasions is
    stable.

    This runs in O(n^4) time, where n is the combination of
    bookings and occasions. So this is a testing tool, not something to
    run in production.

    """

    for attendee in attendees:
        for booking in attendee.accepted:
            for occasion in occasions:

                # the booking was actually accepted, skip
                if booking in occasion.bookings:
                    continue

                over = occasion.preferred(booking)

                if over:
                    for o in occasions:
                        if o != occasion and o.preferred(over):
                            return False

    return True



from datetime import date, timedelta
from functools import partial
from itertools import count
from onegov.activity.matching import MatchableBooking
from onegov.activity.matching import MatchableOccasion
from onegov.activity.matching import deferred_acceptance


today = date.today
keycount = count(start=1, step=1)

match = partial(
    deferred_acceptance,
    stability_check=True,
    validity_check=True,
    hard_budget=True
)


class Booking(MatchableBooking):

    def __init__(self, occasion, attendee, state, priority, start, end):
        self.occasion = occasion
        self.attendee = attendee
        self._id = next(keycount)
        self._state = state
        self._priority = priority
        self._start = start
        self._end = end

    def __eq__(self, other):
        return self._id == other._id

    def __hash__(self):
        return hash(self._id)

    @property
    def id(self):
        return self._id

    @property
    def occasion_id(self):
        return self.occasion

    @property
    def attendee_id(self):
        return self.attendee

    @property
    def score(self):
        return self.priority

    @property
    def state(self):
        return self._state

    @property
    def priority(self):
        return self._priority

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end


class Occasion(MatchableOccasion):

    def __init__(self, name, start, end, max_spots=10):
        self.name = name
        self._max_spots = max_spots
        self.start = start
        self.end = end

    @property
    def id(self):
        return self.name

    @property
    def max_spots(self):
        return self._max_spots

    def booking(self, attendee, state, priority):
        return Booking(
            self.id, attendee, state, priority, self.start, self.end)


def test_overlapping_bookings():

    def days(n):
        return timedelta(days=n)

    # the algorithm will block other bookings, favoring higher priorities
    o1 = Occasion("Daytrip", today(), today())
    o2 = Occasion("Camp", today(), today() + days(1))
    o3 = Occasion("Zoo", today() + days(1), today() + days(2))

    bookings = [
        o1.booking("Justin", 'open', 1),
        o2.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1]}

    # the order of the bookings doesn't matter
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 1)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.blocked == {bookings[0]}
    assert result.accepted == {bookings[1]}

    # if the bookings found are instable, the algorithm will correct that
    bookings = [
        o1.booking("Justin", 'blocked', 1),
        o2.booking("Justin", 'accepted', 0)
    ]

    result = match(bookings, (o1, o2))

    assert not result.open
    assert result.accepted == {bookings[0]}
    assert result.blocked == {bookings[1]}

    # always prefer the booking with the highest priority, even if it leads
    # to more blocked bookings than it would otherwise
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 1),
        o3.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3))

    assert not result.open
    assert result.blocked == {bookings[0], bookings[2]}
    assert result.accepted == {bookings[1]}

    # be predictable if there are no other preferences
    bookings = [
        o1.booking("Justin", 'open', 0),
        o2.booking("Justin", 'open', 0),
        o3.booking("Justin", 'open', 0)
    ]

    result = match(bookings, (o1, o2, o3))

    assert not result.open
    assert result.accepted == {bookings[0], bookings[2]}
    assert result.blocked == {bookings[1]}


def test_accept_highest_priority():

    o = Occasion("Best Activity Ever", today(), today(), max_spots=2)
    bookings = [
        o.booking("Tick", 'open', 0),
        o.booking("Trick", 'open', 1),
        o.booking("Track", 'open', 1)
    ]

    result = match(bookings, [o])

    assert result.open == {bookings[0]}
    assert result.accepted == {bookings[1], bookings[2]}
    assert not result.blocked

    # all other things being equal, the choice is inherently random.
    # to be predictable we need to change all the set and dictionaries to
    # their ordered counterparts which is something we want to avoid for
    # performance/memory usage reasons
    o = Occasion("Best Activity Ever", today(), today(), max_spots=2)
    bookings = [
        o.booking("Tick", 'open', 0),
        o.booking("Trick", 'open', 0),
        o.booking("Track", 'open', 1)
    ]

    result = match(bookings, [o])

    # the first result loses, because it's the first result with a lower
    # score than the third booking - we could turn this around, the important
    # thing is that we are predictable
    assert result.open == {bookings[0]} or result.open == {bookings[1]}
    assert result.accepted == {bookings[1], bookings[2]} or\
        result.accepted == {bookings[0], bookings[2]}
    assert not result.blocked



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""

from onegov.activity import Booking, Period
from onegov.core.orm.types import UUID
from onegov.core.upgrade import upgrade_task
from sqlalchemy import Boolean, Column, Enum, ForeignKey
from sqlalchemy.orm import joinedload


@upgrade_task('Rebuild bookings')
def rebuild_bookings(context):
    # having not created any bookings yet, we can rebuild them
    context.operations.drop_table('bookings')


@upgrade_task('Add period_id to bookings')
def add_period_id_to_bookings(context):

    context.operations.add_column('bookings', Column(
        'period_id', UUID, ForeignKey("periods.id"), nullable=True
    ))

    bookings = context.session.query(Booking)
    bookings = bookings.options(joinedload(Booking.occasion))

    for booking in bookings:
        booking.period_id = booking.occasion.period_id

    context.session.flush()
    context.operations.alter_column('bookings', 'period_id', nullable=False)


@upgrade_task('Change booking states')
def change_booking_states(context):

    new_type = Enum(
        'open',
        'blocked',
        'accepted',
        'denied',
        'cancelled',
        name='booking_state'
    )

    op = context.operations

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state TYPE Text;
        UPDATE bookings SET state = 'open' WHERE state = 'unconfirmed';
        DROP TYPE booking_state;
    """)

    new_type.create(op.get_bind())

    op.execute("""
        ALTER TABLE bookings ALTER COLUMN state
        TYPE booking_state USING state::text::booking_state;
    """)


@upgrade_task('Add confirmed flag to period')
def add_confirmed_flag_to_period(context):
    context.operations.add_column('periods', Column(
        'confirmed', Boolean, nullable=True, default=False
    ))

    for period in context.session.query(Period):
        period.confirmed = False

    context.session.flush()
    context.operations.alter_column('periods', 'confirmed', nullable=False)



import sedate

from onegov.activity import log


def overlaps(booking, other):
    """ Returns true if the two given bookings overlap. """

    assert booking != other

    return sedate.overlaps(
        booking.start, booking.end,
        other.start, other.end
    )


class LoopBudget(object):
    """ Helps ensure that a loop doesn't overreach its complexity budget.

    For example::

        budget = LoopBudget(max_ticks=10)

        while True:
            if budget.limit_reached():
                break
    """

    def __init__(self, max_ticks):
        self.ticks = 0
        self.max_ticks = max_ticks

    def limit_reached(self, as_exception=False):
        self.ticks += 1

        if self.ticks >= self.max_ticks:
            message = "Loop limit of {} has been reached".format(self.ticks)

            if as_exception:
                raise RuntimeError(message)
            else:
                log.warning(message)

            return True


def hashable(attribute):

    class Hashable(object):

        def __hash__(self):
            return hash(getattr(self, attribute))

        def __eq__(self, other):
            return getattr(self, attribute) == getattr(other, attribute)

    return Hashable



from onegov.activity.matching.core import deferred_acceptance
from onegov.activity.matching.core import deferred_acceptance_from_database
from onegov.activity.matching.interfaces import MatchableBooking
from onegov.activity.matching.interfaces import MatchableOccasion

__all__ = [
    'deferred_acceptance',
    'deferred_acceptance_from_database',
    'MatchableBooking',
    'MatchableOccasion'
]



from onegov.activity.models import Booking, Period
from onegov.core.collection import GenericCollection


class BookingCollection(GenericCollection):

    def __init__(self, session, period_id=None, username=None):
        super().__init__(session)
        self.period_id = period_id
        self.username = username

    def query(self):
        query = super().query()

        if self.username is not None:
            query = query.filter(Booking.username == self.username)

        if self.period_id is not None:
            query = query.filter(Booking.period_id == self.period_id)

        return query

    def for_period(self, period):
        return self.__class__(self.session, period.id, self.username)

    def for_username(self, username):
        return self.__class__(self.session, self.period_id, username)

    @property
    def model_class(self):
        return Booking

    def count(self, usernames='*', periods='*', states='*'):
        """ Returns the number of bookings, optionally filtered by usernames,
        periods and states.

        All parameters may either be iterables or subqueries.

        """

        query = self.query().with_entities(Booking.id)

        if states != '*':
            query = query.filter(Booking.state.in_(states))

        if periods != '*':
            query = query.filter(Booking.period_id.in_(periods))

        if usernames != '*':
            query = query.filter(Booking.username.in_(usernames))

        return query.count()

    def wishlist_count(self, username):
        """ Returns the number wishlist entries in the active period.

        This value is equal to the number of bookings for a given user in
        the active period.

        """

        periods = self.session.query(Period)
        periods = periods.with_entities(Period.id)
        periods = periods.filter(Period.active == True)
        periods = periods.filter(Period.confirmed == False)

        return self.count(
            usernames=(username, ),
            periods=periods.subquery(),
            states='*'
        )

    def booking_count(self, username):
        """ Returns the number of accepted bookings in the active period. """

        periods = self.session.query(Period)
        periods = periods.with_entities(Period.id)
        periods = periods.filter(Period.active == True)
        periods = periods.filter(Period.confirmed == True)

        return self.count(
            usernames=(username, ),
            periods=periods.subquery(),
            states=('accepted', )
        )

    def by_user(self, user):
        return self.query().filter(Booking.username == user.username)

    def by_username(self, username):
        return self.query().filter(Booking.username == username)

    def by_occasion(self, occasion):
        return self.query().filter(Booking.occasion_id == occasion.id)

    def add(self, user, attendee, occasion, priority=None, group_code=None):

        return super().add(
            username=user.username,
            attendee_id=attendee.id,
            occasion_id=occasion.id,
            priority=priority,
            group_code=group_code,
            period_id=occasion.period_id
        )
